#!/usr/bin/perl

# Try to keep the programming style good ;-)

use strict;

# By default we don't use the warnings pragma as the system perl
# distribution of FreeBSD 4.x doesn't support the warnings pragma.

#use warnings;

# Some modules we need for password-related Stuff

use Crypt::PasswdMD5;
use Crypt::RandPasswd;

# We want to parse commandline arguments in a comfortable way

use Getopt::Std;

use Config;

# Parse the commandline arguments and fall back to default values if
# neccessary

my %opts;
getopts('b:g:i:l:S:o:m:x:e:D:t:G:hrsdy', \%opts);

if ( $opts{h} ) { usage(); }

sub usage { die "usage: $0 [-i input file] [-g group] [-b base directory for \$HOME] [-s] [-S smbpasswd binary] [-o output file] [-r] [-m minimum password length] [-x maximum password length] [-d] [-D <descriptors for the fields in input file (semicolon separated)> [-y]] [-t <perl regexp that describes the delimiter in the input file>] [-G <member groups>]\n";}

if ( $> != 0 ) {die "You must be root to run this.\n";}

if ( ! scalar %opts ) { usage(); }

my $userliste = "userlist";
if ( $opts{i} ) { $userliste = $opts{i}; }

my $group = "schueler";
if ( $opts{g} ) { $group = $opts{g}; }

if ( ! getgrnam $group ) { die "Group $group doesn't exist. Please create it or specify an existing group.\n" }

my $homebase = "/home/$group";
if ( $opts{b} ) { $homebase = $opts{b}; }

my $smbpass = "/usr/local/bin/smbpasswd";
if ( $opts{S} ) { $smbpass = $opts{S}; }

my $outfile = "useraccounts";
if ( $opts{o} ) { $outfile = $opts{o}; }

my $passtype = "letters";
if ( $opts{r} ) { $passtype = "chars"; }

my @descriptor = ("username","real name","class");
if ( $opts{D} ) { @descriptor = split(/\;/, $opts{D}); }

my $delimiter = '/\t/';
if ( $opts{t} ) { $delimiter = $opts{t}; }

if ( $delimiter !~ /^\/.*\/$/ ) { die "$delimiter is not a valid regular expression. A correct example would be '/\t/'.\n"; }

my $member_groups;
if ( $opts{G} ) { $member_groups = $opts{G}; }

my @member_groups_check;

if ( $member_groups )
	{
	@member_groups_check = split(/,/, $member_groups);

	# Now check if all the member groups specified do exist

	membergroups_check();
	}

sub membergroups_check
{
my @failed_groups;
foreach my $membergroup ( @member_groups_check )
	{
	if ( ! getgrnam $membergroup )
		{
		warn "$membergroup doesn't exist\n";
		push @failed_groups, $membergroup;
		}
	}

# getgrnam returns 0 as group ID of wheel or similiar groups with GID
# zero (of course). Because perl would interpret this as false we need
# this little workarround so the script doesn't die if users are members
# of the wheel group.

if ( @failed_groups )
		{
		foreach my $checkgroup ( @failed_groups )
			{
			if ( (getgrgid 0) eq $checkgroup ) { shift @failed_groups; }
			}
			if ( @failed_groups ) { die "The following groups you specified as member groups for the users to be created do not exist: @failed_groups\n"; }
		}

}

my $member_groups_arg;
if ( $member_groups ) { $member_groups_arg = "-G" . $member_groups; }
			
# Although we require the person running the script to supply the
# enclosing slashes of the regexp we remove them now as the pattern
# won't work otherwise.

$delimiter =~ s/^\/(.*)\/$/$1/;

my $minlength;
if ( $opts{m} ) { $minlength = $opts{m}; }
else	{
	if ( $opts{r} ) { $minlength = 10; }
	else { $minlength = 8; }
	}

my $maxlength;
if ( $opts{x} ) { $maxlength = $opts{x}; }
else	{
	if ( $opts{r} ) { $maxlength = 10; }
	else { $maxlength = 8; }
	}

print "Group: $group\n";
print "Base directory: $homebase\n";
print "Userliste: $userliste\n\n";

# Determine the system we run on

my $os = $Config{osname};

# Create the filehandle to read the userlist from or read from STDIN if
# $opts{i} is "-"

if ( $opts{i} eq "-" ) { open LISTE,"&STDIN"; }
else
	{
	open LISTE,$userliste or die "$userliste: No such file or directory.\n";
	}

# Do some deep magic to determine the school year.

# First get the raw time
my @time = localtime;

# Extract the relevant parts

my $month = $time[4];
my $year = $time[5];

# Convert the year to 2-digit format

my $year1 = sprintf("%02d", $year % 100);

# Now it's time for the arcane part...

my $schoolyear;
if ($month le 7) {
	my $year2=$year1 - 1;
	$schoolyear = "0$year2/$year1";
}
else {
	my $year2=$year1 + 1;
	$schoolyear = "$year1/0$year2";
}

# Make -s implicit if -S is given

if ( $opts{S} and ! $opts{s} )
	{
	$opts{s} = 1;
	}

# Check for smbpasswd(8) if -s is given

if ( $opts{s} ) 
	{
	if ( ! -x $smbpass ) { die "$smbpass doesn't exist or isn't executable. Please specify.\n"; }
	}

# Extract the account data from the list of usernames

while (my $line = <LISTE>)
{
	# Parse the userlist and put it's elements in variables

	my @entry = split(/$delimiter/, $line);

	# First check if @descriptor and @entry contain an equal number
	# of entries

	if ( (scalar @descriptor) != (scalar @entry) ) { die "The number of fields in the input file doesn't match the number of descriptors\n"; }

	my @descriptor_process = @descriptor;

	# shorten @descriptor by one to match @entry
	shift @descriptor_process;
	
	my $name = shift @entry;

	# Remove the newline left over from parsing the input file
	chomp @entry;

	my $output_rest;
	my @entry_process = @entry;
	while ( @entry_process && @descriptor_process)
		{
		$output_rest .= (shift @descriptor_process) . ": " . (shift @entry_process ). " ";
		}
	if ( $opts{y} ) { $output_rest .= $schoolyear };

	$output_rest =~ s/\;$//;

	# Determine if we should try appending random characters to duplicate
	# user names or simply log and report them.

	if ( $opts{d} ) { $name = dupe_engine($name); }
	else { next if dupe_log($name,$output_rest); }

	sub dupe_engine
		{
		# Check if the user exists
		my $username = $_[0];
		if ( getpwnam $username ) 
			{
			print "User $username already in passwd database.\n";
			until ( ! getpwnam $username )
				{
				my $randomizer = Crypt::RandPasswd->word(1,2);
				$username .= $randomizer;
				print "New Name: $username\n";
				}
			}
		return $username;
		}
	sub dupe_log
		{
		# Check if the user exists
		my $username = $_[0];
		my $output_rest = $_[1];
		if ( getpwnam $username )
			{
			my $errlog = "duplicate_users";
			if ( $opts{e} ) { $errlog = $opts{e}; }
			warn "User $username already in passwd database.\n";
			open DUPELOG,">> $errlog";
			print DUPELOG "Username: $username $output_rest\n";
			close DUPELOG;
			}
		else {return;}
		}
	
	# Compose the GECOS field for /etc/passwd

	my $gecos;
	while ( @entry )
		{
		$gecos .= (shift @entry) . ";";
		}

	# Now add the school year to the GECOS field if desired

	if ( $opts{y} ) { $gecos .= $schoolyear; }

	$gecos =~ s/\;$//;

	# Generate a password for the user

	my $clearpass = Crypt::RandPasswd->$passtype( $minlength, $maxlength );

	if ( $opts{r} and ($clearpass =~ /'/) ) { until ( ! $clearpass =~ /'/ ) {$clearpass = Crypt::RandPasswd->$passtype($minlength, $maxlength);} }
	my $passwdhash = unix_md5_crypt ($clearpass, '$1$');

	# Create an Unix account
	# First determine which sytem we run on and which commands we
	# have to use

	if ( $os eq "linux" )
		{
		# useradd(8) is braindead because it won't let me
		# specify a base directory for $HOME
		my $home = "$homebase/$name";
		if ( $home =~ /\/{2,}/ ) { $home =~ s/\/{2,}/\//g; }

		my @useradd_cmd = ( "/usr/sbin/useradd", "-m", "-c$gecos", "-g$group", "-d$home" );
		if ( $member_groups_arg ) { push @useradd_cmd, $member_groups_arg; }
		push @useradd_cmd, $name;
		system @useradd_cmd;
		open PASSPIPE,"| /usr/sbin/chpasswd -e";
		print PASSPIPE "$name:$passwdhash\n";
		close PASSPIPE;
		}

	elsif ( $os eq "freebsd" )
		{
		my @pw_args = ("useradd", "-n $name", "-m", "-c \'$gecos\'", "-g $group", "-b $homebase", "-h 0");
		if ( $member_groups_arg ) { push @pw_args, $member_groups_arg; } 
		open PWPIPE, "| /usr/sbin/pw @pw_args";
		print PWPIPE "$clearpass\n";
		close PWPIPE;
		}

	else { die "Your OS is not supported\n"; }


	# Use smbpasswd(8) to create an smbpasswd entry
	# Yes this is ugly...and so is smbpasswd's non-interactive
	# mode...

	# Compose the complete smbpasswd command

	if ( $opts{s} ) { smb_create($name, $clearpass); }

	sub smb_create 
		{
		my $name = $_[0];
		my $clearpass = $_[1];
		my $smbpasswd = "$smbpass -s -a $name";
		open SMBPIPE,"| $smbpasswd";
		print SMBPIPE "$clearpass\n$clearpass\n";
		close SMBPIPE;
		}

	# Write the account data to a file

	open OUTFILE,">>$outfile";
	print OUTFILE "Username: $name password: $clearpass $output_rest\n";
	print "Username: $name password: $clearpass $output_rest\n";
	
}


close LISTE;
close OUTFILE;

chmod (0600, $outfile);
